@@ -158,37 +158,54 @@ enaml wish list
 """
 from abc import abstractmethod, ABC
 import ast
+from copy import deepcopy
 import os
 import pathlib
-from typing import Any, Dict, List, Optional, Sequence
+from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union
 
+from ae.base import norm_line_sep, norm_name                                    # type: ignore
 from ae.core import DEBUG_LEVEL_VERBOSE                                         # type: ignore
 from ae.i18n import get_f_string                                                # type: ignore
 from ae.gui_app import id_of_flow, flow_action, flow_key, replace_flow_action   # type: ignore
 
 
-__version__ = '0.1.20'
+__version__ = '0.1.21'
 
 
+FLOW_PATH_ROOT_ID = 'ROOT'  #: pseudo item id needed for flow path jumper and for drop onto leave item button
+FLOW_PATH_TEXT_SEP = " / "  #: flow path separator for :meth:`~LiszDataMixin.flow_path_text`
+
+FOCUS_FLOW_PREFIX = "->"    #: prefix shown in front of flow key of focused item
+
+INVALID_ITEM_ID_PREFIX_CHARS = '[' + '{'  #: invalid initial chars in item id (for to detect id | literal in flow key)
+# currently only the '[' char is used (for to put the node list data as literal in a flow key - see
+
 NODE_FILE_PREFIX = 'node_'  #: file name prefix for node imports/exports
 NODE_FILE_EXT = '.txt'      #: file extension for node imports/exports
 
-FLOW_PATH_ROOT_ID = 'ROOT'  #: needed for flow path jumper and for drop onto leave item button
-FLOW_PATH_TEXT_SEP = " / "  #: flow path separator for :meth:`~LiszDataMixin.flow_path_text` and drop enter
+
+LiszItem = Dict[str, Any]   #: node item data (nid) type
+LiszNode = List[LiszItem]   #: node/list type
 
 
-LiszItem = Dict[str, Any]           #: node item data (nid) type
-LiszNode = List[LiszItem]           #: node/list type
-
-
-def check_item_id(item_id: str) -> bool:
-    """ check if passed item id is valid.
+def check_item_id(item_id: str) -> str:
+    """ check if the passed item id string is valid.
 
     :param item_id:     item id to check.
-    :return:            True if all characters in the pass item id are valid, else False.
+    :return:            "" if all chars in the passed item id are valid, else one of the translated message strings.
     """
-    return isinstance(item_id, str) and bool(item_id) \
-        and item_id != FLOW_PATH_ROOT_ID and FLOW_PATH_TEXT_SEP not in item_id
+    msg = get_f_string("item id '{item_id}' ")
+    if not isinstance(item_id, str):
+        return msg + get_f_string("has to be a string but got {type(item_id)}")
+    if not item_id.strip():
+        return msg + get_f_string("has to be non-empty")
+    if item_id == FLOW_PATH_ROOT_ID:
+        return msg + get_f_string("cannot be equal to '{FLOW_PATH_ROOT_ID}'")
+    if FLOW_PATH_TEXT_SEP in item_id:
+        return msg + get_f_string("cannot contain '{FLOW_PATH_TEXT_SEP}'")
+    if item_id[0] in INVALID_ITEM_ID_PREFIX_CHARS:
+        return msg + get_f_string("cannot start with one of the characters '{INVALID_ITEM_ID_PREFIX_CHARS}'")
+    return ""
 
 
 def correct_item_id(item_id: str) -> str:
@@ -200,7 +217,25 @@ def correct_item_id(item_id: str) -> str:
     item_id = item_id.replace(FLOW_PATH_TEXT_SEP, '/').strip().strip('@~*-#.,;:').strip()
     if item_id == FLOW_PATH_ROOT_ID:
         item_id += '!'
-    return item_id
+    return item_id.lstrip(INVALID_ITEM_ID_PREFIX_CHARS)
+
+
+def item_sel_filter(item: LiszItem) -> bool:
+    """ callable for to filter selected LiszItems.
+
+    :param item:        item data structure to check.
+    :return:            True if item is selected.
+    """
+    return bool(item.get('sel', False)) is True
+
+
+def item_unsel_filter(item: LiszItem) -> bool:
+    """ callable for to filter selected LiszItems.
+
+    :param item:        item data structure to check.
+    :return:            True if item is not selected.
+    """
+    return bool(item.get('sel', False)) is False
 
 
 class LiszDataMixin(ABC):
@@ -234,9 +269,13 @@ class LiszDataMixin(ABC):
         """ change/switch flow id. """
 
     @abstractmethod
-    def flow_path_action(self, path_index: int = -1) -> str:
+    def flow_path_action(self, flow_path: Optional[List[str]] = None, path_index: int = -1) -> str:
         """ determine the action of the last/newest entry in the flow_path (ae.gui_app.MainAppBase method). """
 
+    @abstractmethod
+    def flow_path_strip(self, flow_path: List[str]) -> List[str]:
+        """ return flow_path copy with all non-enter actions stripped from the end. """
+
     @abstractmethod
     def play_sound(self, sound_name: str):
         """ play audio/sound file (ae.gui_app.MainAppBase method). """
@@ -265,7 +304,7 @@ class LiszDataMixin(ABC):
         """ add item to currently displayed node.
 
         :param items:           LiszNode list to add (each item has to have a non-empty item id).
-        :return:                error message if any error happened (multiple error messages are separated by \n),
+        :return:                error message if any error happened (multiple error messages are separated by \\\\n),
                                 else empty string.
         """
         errors = list()
@@ -300,7 +339,7 @@ class LiszDataMixin(ABC):
         elif lit.startswith('{'):
             items = [ast.literal_eval(lit)]
         else:
-            items = [dict(id=item_id) for item_id in lit.replace('\r\n', '\n').replace('\r', '\n').split('\n')
+            items = [dict(id=item_id) for item_id in norm_line_sep(lit).split('\n')
                      if correct_item_id(item_id)]
         return items
 
@@ -354,9 +393,10 @@ class LiszDataMixin(ABC):
 
             if add_item:
                 new_id = correct_item_id(new_id)
-            elif not check_item_id(new_id):
-                return get_f_string(
-                    "invalid item id '{new_id}' containing '{FLOW_PATH_TEXT_SEP}' or is equal '{FLOW_PATH_ROOT_ID}'")
+            else:
+                chk_err = check_item_id(new_id)
+                if chk_err:
+                    return chk_err
 
             new_idx = self.find_item_index(new_id)
             if new_idx != -1 and (add_item or new_idx != old_idx):
@@ -386,21 +426,26 @@ class LiszDataMixin(ABC):
 
         return ""
 
-    def export_node(self, flow_path: List[str], file_path: str = ".") -> Optional[Exception]:
+    def export_node(self, flow_path: List[str], file_path: str = ".", node: Optional[LiszNode] = None
+                    ) -> Optional[Exception]:
         """ export node specified by the passed :paramref:`~export_node.flow_path` argument.
 
         :param flow_path:   flow path of the node to export.
         :param file_path:   folder to store the node data into (def=current working directory).
+        :param node:        explicit/filtered node items (if not passed then all items will be exported).
         :return:            `None` if node got exported without errors, else the raised exception.
         """
-        node = self.flow_path_node(flow_path)
-        file_name = NODE_FILE_PREFIX + (flow_key(flow_path[-1]) if flow_path else FLOW_PATH_ROOT_ID) + NODE_FILE_EXT
+        if not node:
+            node = self.flow_path_node(flow_path)
+        flow_path = self.flow_path_strip(flow_path)
+        file_name = NODE_FILE_PREFIX + (norm_name(flow_key(flow_path[-1])) if flow_path else
+                                        FLOW_PATH_ROOT_ID) + NODE_FILE_EXT
 
         try:
-            # the alternative `os.makedirs(exist_ok=True)` has problems on POSIX with '..'
+            # the alternative `os.makedirs(exist_ok=True)` has problems on POSIX with '..' in the path
             pathlib.Path(file_path).mkdir(parents=True, exist_ok=True)
             with open(os.path.join(file_path, file_name), 'w') as file_handle:
-                file_handle.write(str(node))
+                file_handle.write(repr(node))
         except (FileExistsError, FileNotFoundError, OSError, ValueError) as ex:
             return ex
         return None
@@ -439,7 +484,7 @@ class LiszDataMixin(ABC):
                 if item_id != key:
                     while item_id.startswith(key[:id_len]) and id_len < key_len:
                         id_len += 1
-            return f" ->{key[:id_len]}"
+            return f" {FOCUS_FLOW_PREFIX}{key[:id_len]}"
         return f".{flow_id}" if self.debug_level >= DEBUG_LEVEL_VERBOSE else ""
 
     def flow_path_from_text(self, text: str) -> List[str]:
@@ -449,7 +494,7 @@ class LiszDataMixin(ABC):
         :return:                flow path list.
         """
         flow_path = list()
-        if text != FLOW_PATH_ROOT_ID:
+        if text not in ('', FLOW_PATH_ROOT_ID):
             node = self.root_node
             for part in text.split(FLOW_PATH_TEXT_SEP):
                 for nid in node:
@@ -463,23 +508,25 @@ class LiszDataMixin(ABC):
                 node = path_nid['node']         # type: ignore   # mypy not supports recursive types see issue #731
         return flow_path
 
-    def flow_path_node(self, flow_path: List[str] = None) -> LiszNode:
-        """ determine the current node specified by the passed flow path.
+    def flow_path_node(self, flow_path: List[str] = None, strict: bool = False) -> LiszNode:
+        """ determine the node specified by the passed flow path.
 
         :param flow_path:       flow path list.
-        :return:                current node list.
+        :param strict:          pass True to return empty list on invalid/broken flow_path.
+        :return:                node list.
         """
         if flow_path is None:
             flow_path = self.flow_path
         current_node_items = self.root_node
         for flow_id in flow_path:
-            if flow_action(flow_id) != 'enter':
-                break
-            node_id = flow_key(flow_id)
-            item = self.item_by_id(node_id, searched_node=current_node_items)
-            if 'node' not in item:
-                break                           # actually not needed, will repair broken flow path
-            current_node_items = item['node']
+            if flow_action(flow_id) == 'enter':
+                node_id = flow_key(flow_id)
+                item = self.item_by_id(node_id, searched_node=current_node_items)
+                if 'node' not in item:
+                    if strict:
+                        current_node_items = list()
+                    break                           # actually not needed, will repair broken flow path
+                current_node_items = item['node']
         return current_node_items
 
     def flow_path_quick_jump_nodes(self) -> List[str]:
@@ -496,22 +543,26 @@ class LiszDataMixin(ABC):
                     append_nodes(nid['node'])
                     deeper_flow_path.pop()
 
-        paths = [FLOW_PATH_ROOT_ID] if self.flow_path_action(0) == 'enter' else list()
-        deep = 0
-        while deep < len(self.flow_path) - 1 and self.flow_path_action(deep + 1) == 'enter':
-            deep += 1
-            paths.append(self.flow_path_text(self.flow_path[:deep]))
+        paths = [FLOW_PATH_ROOT_ID] if self.flow_path_action(path_index=0) == 'enter' else list()
 
-        deeper_flow_path = self.flow_path[:-1]      # start at current flow path minus the open_flow_path_jumper flow id
+        # add flow path parents up-to/without-already added root
+        deep = 1
+        while deep < len(self.flow_path) and self.flow_path_action(path_index=deep) == 'enter':
+            paths.append(self.flow_path_text(self.flow_path[:deep]))
+            deep += 1
+
+        # add sub paths, find start node from the end of current flow path, skipping opt. open_flow_path_jumper flow id
+        deeper_flow_path = self.flow_path_strip(self.flow_path)
         append_nodes(self.current_node_items)
 
         return paths
 
-    def flow_path_text(self, flow_path: List[str], min_len: int = 3) -> str:
+    def flow_path_text(self, flow_path: List[str], min_len: int = 3, display_root: bool = False) -> str:
         """ generate shortened display text from the passed flow path.
 
         :param flow_path:       flow path list.
         :param min_len:         minimum length of node ids (flow id keys).
+        :param display_root:    pass True to return FLOW_PATH_ROOT_ID on empty/root path.
         :return:                shortened display text string of the passed flow path (which can be converted back
                                 to a flow path list with the method :meth:`.flow_path_from_text`).
         """
@@ -535,7 +586,7 @@ class LiszDataMixin(ABC):
             if path_nid and 'node' in path_nid:     # prevent error in quick jump to root
                 node = path_nid['node']             # type: ignore   # mypy not supports recursive types see issue #731
 
-        return FLOW_PATH_TEXT_SEP.join(shortened_ids)
+        return FLOW_PATH_TEXT_SEP.join(shortened_ids) if shortened_ids else (FLOW_PATH_ROOT_ID if display_root else '')
 
     def focus_neighbour_item(self, delta: int):
         """ move flow id to previous/next displayed/filtered node item.
@@ -625,6 +676,45 @@ class LiszDataMixin(ABC):
 
         return True
 
+    def node_info(self, flow_path: List[str], what: Tuple[str, ...] = ()) -> Dict[str, Union[int, str, List[str]]]:
+        """ determine statistics info for the node specified by :paramref:`~node_info.flow_path`.
+
+        :param flow_path:       flow path to identify node (pass empty list to get the info of the root node).
+        :param what:            pass tuple of statistic info fields for to include only these into the returned dict
+                                (passing an empty tuple or nothing will include all the following fields):
+
+                                * 'count': number of items (nodes and leaves) in this node (including sub-nodes).
+                                * 'leaf_count': number of sub-leaves.
+                                * 'node_count': number of sub-nodes.
+                                * 'selected_leaf_count': number of selected sub-leaves.
+                                * 'unselected_leaf_count': number of unselected sub-leaves.
+                                * 'name': name of the item containing this node.
+                                * 'names': list of all sub-item/-node names/ids.
+                                * 'leaf_names': list of all sub-leaf names.
+                                * 'selected_leaf_names': list of all selected sub-leaf names.
+                                * 'unselected_leaf_names': list of all unselected sub-leaf names.
+
+        :return:                dict with the node info specified by the :paramref:`~node_info.what` argument.
+        """
+        if flow_path:
+            parent_node = self.flow_path_node(flow_path[:-1], strict=True)
+            name = flow_key(flow_path[-1])
+        else:
+            parent_node = self.root_node
+            name = FLOW_PATH_ROOT_ID
+
+        names = self.sub_item_ids(node=parent_node, item_id=name, leaves_only=False)
+        count = len(names)
+        leaf_names = self.sub_item_ids(node=parent_node, item_id=name)
+        leaf_count = len(leaf_names)
+        selected_leaf_names = self.sub_item_ids(node=parent_node, item_id=name, hide_sel_val=False)
+        selected_leaf_count = len(selected_leaf_names)          # noqa: F841
+        unselected_leaf_names = self.sub_item_ids(node=parent_node, item_id=name, hide_sel_val=True)
+        unselected_leaf_count = len(unselected_leaf_names)      # noqa: F841
+        node_count = count - leaf_count                         # noqa: F841
+
+        return {k: v for k, v in locals().items() if not what or k in what}
+
     def on_app_init(self):
         """ initialize self.current_node_items from flow_id/flow_path saved on last app stop. """
         self.refresh_current_node_items_from_flow_path()
@@ -772,6 +862,50 @@ class LiszDataMixin(ABC):
         event_kwargs['flow_id'] = id_of_flow('focus', 'item', item_id)
         return True
 
+    def on_node_extract(self, flow_path_text: str, event_kwargs: dict) -> bool:
+        """ FlowButton tap event handler for to extract (copy/cut/delete/export) the node specified by `flow_path_text`.
+
+        :param flow_path_text:  flow path text or list literal (identifying the node to extract).
+        :param event_kwargs:    event arguments (used for to pass the `export_path` and `extract_type`):
+
+                                `export_path` specifies the destination folder of the export (default='.'/CWD).
+
+                                `extract_type` specifies extract destination and an optional filter on un-/selected
+                                items, e.g. the following string values can be passed for a 'copy' destination extract:
+
+                                * 'copy' is copying all items of the specified node to the clipboard.
+                                * 'copy_sel' is copying only the selected items of the node to the clipboard.
+                                * 'copy_unsel' is copying only the unselected items to the clipboard.
+
+        :return:                True for to process/change flow.
+        """
+        flow_path = ast.literal_eval(flow_path_text) if flow_path_text and flow_path_text[0] == '[' else \
+            self.flow_path_from_text(flow_path_text)
+        node = self.flow_path_node(flow_path)
+        dst, *what = event_kwargs['extract_type'].split('_')
+
+        if not what:
+            extract_filter = None
+            delete_filter = lambda item: True   # noqa: E731
+        elif what[0] == 'sel':
+            extract_filter = item_unsel_filter
+            delete_filter = item_sel_filter
+        else:
+            extract_filter = item_sel_filter
+            delete_filter = item_unsel_filter
+
+        extract_node = deepcopy(node)
+        if dst in ('cut', 'delete'):
+            self.shrink_node_size(node, item_filter=delete_filter)      # in-place deletion
+        self.shrink_node_size(extract_node, item_filter=extract_filter)
+
+        if dst in ('copy', 'cut'):
+            self.call_method('on_clipboard_key_c', repr(extract_node))  # Clipboard.copy(repr(node))
+        elif dst == 'export':
+            self.export_node(flow_path, file_path=event_kwargs.get('export_path', '.'), node=extract_node)
+
+        return True
+
     def on_node_jump(self, flow_path_text: str, event_kwargs: dict) -> bool:
         """ FlowButton clicked event handler restoring flow path from the flow key.
 
@@ -789,7 +923,8 @@ class LiszDataMixin(ABC):
         # here (for it can be removed by FlowPopup closed event handler when the jumper popup closes).
         # open_jumper_flow_id = id_of_flow('open', 'flow_path_jumper')
         # assert open_jumper_flow_id == self.flow_path[-1]
-        flow_path.append(self.flow_path[-1])
+        if self.flow_path_action(flow_path) == 'enter' and self.flow_path_action() == 'open':
+            flow_path.append(self.flow_path[-1])
 
         self.change_app_state('flow_path', flow_path, send_event=False)
         event_kwargs['flow_id'] = id_of_flow('', '')
@@ -829,51 +964,61 @@ class LiszDataMixin(ABC):
         """ refresh current node including the depending display node. """
         self.current_node_items = self.flow_path_node()
 
-    def shrink_node_size(self, node: LiszNode):
-        """ shrink size of node and sub-node keys by removing unneeded `sel` keys, for to save space in config file.
+    def shrink_node_size(self, node: LiszNode, item_filter: Optional[Callable[[LiszItem], bool]] = None):
+        """ shrink node size by removing unneeded items and `sel` keys, e.g. for to save space in config file.
 
-        :param node:            start or root node to shrink.
+        :param node:            start or root node to shrink (in-place!).
+        :param item_filter:     pass callable for to remove items from the passed node and its sub-nodes.
+                                The callable is getting passed each item as argument and has to return True
+                                for to remove it from its node.
         """
+
+        del_items = list()
         for item in node:
-            if item.get('sel', 1) != 1:
+            is_node = 'node' in item
+            if not is_node and item_filter and item_filter(item):
+                del_items.append(item)
+            elif item.get('sel', 1) != 1:
                 item.pop('sel')
             elif 'sel' in item:
                 item['sel'] = 1  # remove also decimal point and zero (float to int)
-            if 'node' in item:
-                self.shrink_node_size(item['node'])
+            if is_node:
+                self.shrink_node_size(item['node'], item_filter=item_filter)
+        for item in del_items:
+            node.remove(item)
 
-    def sub_item_ids(self, item_id: str, leaves_only: bool, hide_sel_val: Optional[bool] = None,
-                     node: Optional[LiszNode] = None, sub_ids: Optional[List[str]] = None) -> List[str]:
+    def sub_item_ids(self, node: Optional[LiszNode] = None, item_id: str = '',
+                     leaves_only: bool = True, hide_sel_val: Optional[bool] = None,
+                     sub_ids: Optional[List[str]] = None) -> List[str]:
         """ return item names of all items including their sub-node items (if exists).
 
         Used for to determine the affected items if user want to delete or de-/select the sub-items of
         the node item specified by the :paramref:`sub_item_ids.item_id` argument.
 
+        :param node:            searched node, if not passed use root node if the :paramref:`~sub_item_ids.item_id`
+                                argument is :data:`FLOW_PATH_ROOT_ID` else use the current node as default.
         :param item_id:         item id of the node item to be investigated (for to be deleted or de-/selected later).
-        :param leaves_only:     True if only include the leaf item ids of the sub-nodes of the node item.
+                                Pass FLOW_PATH_ROOT_ID or '' for to get the item names/ids within AND underneath
+                                of the node specified by :paramref:`~sub_item_ids.node` (or its default value).
+        :param leaves_only:     True (the default) if only include the leaf item ids of the sub-nodes of the node item.
         :param hide_sel_val:    pass False/True for to exclude un-/selected items from the returned list of ids.
                                 If None or not passed then all found items will be included.
-        :param node:            searched node, use current node if not passed (used for recursive calls of this method).
         :param sub_ids:         already found sub item ids (used only for the recursive calls of this method).
         :return:                list of found item ids.
         """
         if node is None:
-            node = self.current_node_items
+            node = self.root_node if item_id == FLOW_PATH_ROOT_ID else self.current_node_items
+        if item_id not in ('', FLOW_PATH_ROOT_ID):
+            node = self.item_by_id(item_id, searched_node=node).get('node', list())
         if sub_ids is None:
             sub_ids = list()
 
-        item = self.item_by_id(item_id, node)
-
-        if not leaves_only and (hide_sel_val is None or bool(item.get('sel')) != hide_sel_val):
-            sub_ids.append(item_id)
-
-        node = item.get('node')
-        if node is not None:
-            for item in node:
-                if item.get('node'):
-                    self.sub_item_ids(item['id'], leaves_only, hide_sel_val=hide_sel_val, node=node, sub_ids=sub_ids)
-                elif hide_sel_val is None or bool(item.get('sel')) != hide_sel_val:
-                    sub_ids.append(item['id'])
+        for item in node:           # type: ignore # mypy does not recognize that node cannot be None
+            if not leaves_only if 'node' in item else (hide_sel_val is None or bool(item.get('sel')) != hide_sel_val):
+                sub_ids.append(item['id'])
+            if item.get('node'):
+                self.sub_item_ids(node=node, item_id=item['id'], leaves_only=leaves_only, hide_sel_val=hide_sel_val,
+                                  sub_ids=sub_ids)
 
         return sub_ids
 
@@ -882,8 +1027,7 @@ class LiszDataMixin(ABC):
 
         :param node_idx:            list index of the item in the current node to change the selection for.
         """
-        selected = self.current_node_items[node_idx].get('sel', False)
-        if selected:
+        if item_sel_filter(self.current_node_items[node_idx]):
             self.current_node_items[node_idx].pop('sel')
         else:
             self.current_node_items[node_idx]['sel'] = 1.0
